/*
 *  -*- Mode: ANSI C -*-
 *  $Id: flwt.c,v 1.18 1997/04/03 15:08:04 fernande Exp $
 *  $Source: /sgi.acct/sweldens/cvs/liftpack/Lifting/flwt.c,v $
 *  Author: Gabriel Fernandez, Senthil Periaswamy
 *
 *     >>> Fast Lifted Wavelet Transform on the Interval <<<
 *
 *                             .-. 
 *                            / | \
 *                     .-.   / -+- \   .-.
 *                        `-'   |   `-'
 *
 *  Using the "Lifting Scheme" and the "Square 2D" method, the coefficients
 *  of a "Biorthogonal Wavelet" or "Second Generation Wavelet" are found for
 *  a two-dimensional data set. The same scheme is used to apply the inverse
 *  operation.
 */
/* do not edit anything above this line */

/* System header files */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>   /* exit() */
/* FLWT header files */
#include <constant.h>
#include <flwterr.h>
#include <filter.h>
#include <flwt.h>
#include <imgmem.h>
#include <lift.h>
#include <mallat.h>
#include <mem.h>
#include <util.h>

/* Static Global Variables */
static Vector filter = NULL;      /* Vector with filter coefficients */
static Matrix liftingX = NULL;    /* Matrix with lifting coefficients in X */
static Matrix liftingY = NULL;    /* Matrix with lifting coefficients in Y */

/* code */

/* CODE FOR THE FILTER AND LIFTING COEFFICIENTS */

/*
 * GetDual function: Finds the N filter coefficients and allocates a global
 *                   vector with the values to be used in the transform.
 *                   Allocates the global lifting matrices and initialize the
 *                   integral-moments information to calculate the lifting
 *                   coefficients.
 *                   These filter coefficients are used to find the Dual
 *                   Wavelet function.
 */
void
GetDual ( const int width, const int height, const int N, const int nTilde,
          const boolean print, const int levels )
{
    int row, col, noGammas, nX, nY;
    Flt maxN;
    Vector filterPtr;

    /* Calculate Filter Coefficients */
    if ( (filter = GetFilter(N)) == NULL ) {
        Error ("GetDual", NOT_EVEN, RETURN, "N");
        Error ("GetDual", INVALID_RELATIONAL_INT, ABORT, "N", ">", 0);
    } else {
        if (print) {
            int Nrows = (N>>1) + 1;
            filterPtr = filter;
            fprintf (stdout, "\nFilter Coefficients (N=%d)\n", N);
            for (row = 0 ; row < Nrows ; row++) {
                fprintf (stdout, "For %d in the left and %d in the right\n[ ",
                         row, N-row);
                for (col = 0 ; col < N ; col++)
                    fprintf (stdout, "%.18g ", *(filterPtr++));
                fprintf (stdout, "]\n");
            }
        }
    }

    /* Allocate lifting coefficient matrices */
    maxN = (Flt)MAX(N, nTilde) - (Flt)1;   /* max vanishing moments */
    noGammas = width >> 1;    /* number of Gammas in X direction */
    if (noGammas > 0) {
        nX = (int)logBaseN ( (Flt)(width-1)/maxN, (Flt)2 );   /* iterations in X */
        nX = (nX < 0) ? 0 : MIN(levels, nX);   /* find lowest level */
        liftingX = matrix( 0, (long)(nX-1), 0, (long)(noGammas*nTilde - 1) );
        for ( row=0 ; row<nX ; row++ )
            for ( col=0 ; col<noGammas*nTilde ; col++ )
                liftingX[row][col] = (Flt)0;
    }
    noGammas = height >> 1;    /* number of Gammas in Y direction */
    if (noGammas > 0) {
        nY = (int)logBaseN ( (Flt)(height-1)/maxN, (Flt)2 );   /* iterations in Y */
        nY = (nY < 0) ? 0 : MIN(levels, nY);   /* find lowest level */
        liftingY = matrix( 0, (long)(nY-1), 0, (long)(noGammas*nTilde - 1) );
        for ( row=0 ; row<nY ; row++ )
            for ( col=0 ; col<noGammas*nTilde ; col++ )
                liftingY[row][col] = (Flt)0;
    }

    /* Initialize Integral-Moments matrices */
    if (!InitMoment (nTilde, width, height, print)) {
        Error ("GetDual", NOT_EVEN, RETURN, "N");
        Error ("GetDual", INVALID_RELATIONAL_INT, ABORT, "N", ">", 0);
    }
}

/*
 * FreeCo function: Frees the memory allocated by the function GetDual().
 */
void
FreeCo ( const int width, const int height, const int N, const int nTilde,
         const int levels )
{
    int noGammas, nRows, nX, nY;
    Flt maxN;

    /* Free Filter Coefficients Vector */
    nRows = (N>>1) + 1;
    free_vector( filter, 0, (long)(nRows*N - 1) );

    /* Free Lifting Coefficient Matrices */
    maxN = (Flt)MAX(N, nTilde) - (Flt)1;   /* max vanishing moments */
    if (liftingX) {
        noGammas = width >> 1;
        nX = (int)logBaseN( (Flt)(width-1)/maxN, (Flt)2 );   /* iterations in X */
        nX = (nX < 0) ? 0 : MIN(levels, nX);   /* find lowest level */
        free_matrix( liftingX, 0, (long)(nX-1), 0, (long)(noGammas*nTilde - 1) );
    }
    if (liftingY) {
        noGammas = height >> 1;
        nY = (int)logBaseN ( (Flt)(height-1)/maxN, (Flt)2 );   /* Iterations in Y */
        nY = (nY < 0) ? 0 : MIN(levels, nY);   /* find lowest level */
        free_matrix( liftingY, 0, (long)(nY-1), 0, (long)(noGammas*nTilde - 1) );
    }

    /* Free Integral-Moments Matrices */
    FreeMoment( width, height, nTilde );
}

/*
 * GetReal function: Calculates the moments and lifting coefficients
 *                   needed for the wavelet transform.
 *                   These coefficients preserve the moments of the Real
 *                   Wavelet function.
 */
void
GetReal ( const int width, const int height, const int N, const int nTilde,
          const boolean print, const int levels )
{
    int x, y, k, n, nX, nY, step, max_step, level;
    Flt  maxN;
    Vector liftPtr;
    char buf[BUFSIZ];

    /* Check for filter vector */
    if (!filter) {
        Error ("GetReal", NO_FILTER_VECTOR, ABORT);
    }

    /* Calculate number of iterations n. It is the maximum  */
    /* value such that the following relation is satisfied. */
    /*        (2^n)*(N-1) <= L < (2^(n+1))*(N-1) + 1        */
    /* Where L = max (signal's length in X-Y direction).    */
    /* and N = max (# dual vanish mom & # real vanish mom)  */
    /* Hence, solving for n, we have the following equation */
    /* for all the cases:  n = floor (log_2((L-1)/(N-1))    */
    maxN = (Flt)MAX(N, nTilde) - (Flt)1;   /* max vanishing moments */
    /* Iterations in X */
    nX = (width == 1) ? 0 : (int)logBaseN ( (Flt)(width-1)/maxN, (Flt)2 );
    nX = (nX < 0) ? 0 : MIN (levels, nX);   /* find lowest level */
    /* Iterations in Y */
    nY = (height == 1) ? 0 : (int)logBaseN ( (Flt)(height-1)/maxN, (Flt)2 );
    nY = (nY < 0) ? 0 : MIN (levels, nY);   /* find lowest level */
    /* Total number of iterations */
    n = MAX(nX, nY);   /* find minimum number of iterations */
    max_step = (n==0) ? 0 : 1<<n;   /* maximum step */

    /* Find lifting coefficients for all levels */
    level = 0;
    for ( step = 1 ; step < max_step ; step <<= 1 ) {
        if (nX-- > 0) {
            /* Get lifting coefficients for rows */
            GetLifting (liftingX[nX], filter, step, N, nTilde, width, FALSE, FALSE);
            if (print) {
                fprintf (stdout, "\nLifting Coefficients for Rows (Level=%d)\n",
                         level);
                liftPtr = liftingX[nX];
                for ( x=0 ; x<(width>>1) ; x++ ) {
                    sprintf (buf, "%.20g", *(liftPtr++));
                    for ( k=1 ; k<nTilde ; k++ )
                        sprintf (buf, "%s,%.20g", buf, *(liftPtr++));
                    fprintf (stdout, "Gamma[%d] (%s)\n", x, buf);
                }
            }
        }

        if (nY-- > 0) {
            /* Get lifting coefficients for columns */
            GetLifting (liftingY[nY], filter, step, N, nTilde, height, TRUE, FALSE);
            if (print) {
                fprintf (stdout,
                         "\nLifting Coefficients for Columns (Level=%d)\n",
                         level);
                liftPtr = liftingY[nY];
                for ( y=0 ; y<(height>>1) ; y++ ) {
                    sprintf (buf, "%.20g", *(liftPtr++));
                    for ( k=1 ; k<nTilde ; k++ )
                        sprintf (buf, "%s,%.20g", buf, *(liftPtr++));
                    fprintf (stdout, "Gamma[%d] (%s)\n", y, buf);
                }
            }
        }
        level++;   /* go to next level */
    }
}

/* CODE FOR THE IN-PLACE FLWT (INTERPOLATION CASE, N EVEN) */

/*
 * FLWT2D function: This is a 2D Direct/Inverse Fast Lifted Wavelet 
 *                  Trasform that uses the "Square 2D" method to find 
 *                  the Gamma coefficients of the wavelet function Phi
 *                  in the Direct case and the Lambda coefficients of the
 *                  original signal in the Inverse case. The "Square 2D"
 *                  method applies the transform first to all the rows
 *                  and then to all the columns. This creates squared
 *                  window transforms for the process. Since the Lifting
 *                  Scheme is used, the final coefficients are organized
 *                  in-place in the original matrix Data[][]. In file
 *                  mallat.c there are functions provided to change this
 *                  organization to the Isotropic format originally
 *                  established by Mallat.
 */
void
FLWT2D ( Matrix Data, const int width, const int height,
                      const int N, const int nTilde,
                      const int levels, const boolean inverse )
{
    int i, k, x, y, /* counters */
        n,          /* maximum number of levels */
        nX, nY,     /* number of iterations in X and Y directions */
        step,       /* step size of current level */
        max_step;   /* maximum and initial step for the direct and */
                    /* inverse transforms, respectively */
    Flt maxN;       /* maximum number of vanishing moments */


    /* Verify for existence of filter and lifting coefficients */
    if (!filter) {
        Error ("FLWT2D", NO_FILTER_VECTOR, ABORT);
    }

    if ( !liftingX && !liftingY ) {
	fprintf( stderr,
	         "\nFLWT2D(): lifting coefficients are not initialized."
		 "\nUse GetDual() and GetReal() before calling this function.\n");
        exit( INPUT_ERROR );
    }

    /* Calculate number of iterations n. It is the maximum  */
    /* value such that the following relation is satisfied. */
    /*        (2^n)*(N-1) <= L < (2^(n+1))*(N-1) + 1        */
    /* Where L = max (signal's length in X-Y direction).    */
    /* and N = max (# dual vanish mom & # real vanish mom)  */
    /* Hence, solving for n, we have the following equation */
    /* for all the cases:  n = floor (log_2((L-1)/(N-1))    */
    maxN = (Flt)MAX(N, nTilde) - (Flt)1;   /* max vanishing moments */
    /* Iterations in X */
    nX = (width == 1) ? 0 : (int)logBaseN ( (Flt)(width-1)/maxN, (Flt)2 );
    nX = (nX < 0) ? 0 : MIN(levels, nX);   /* find lowest level */
    /* Iterations in Y */
    nY = (height == 1) ? 0 : (int)logBaseN ( (Flt)(height-1)/maxN, (Flt)2 );
    nY = (nY < 0) ? 0 : MIN(levels, nY);   /* find lowest level */
    /* Total number of iterations */
    n = MAX(nX, nY);   /* find minimum number of iterations */


    if ( !inverse ) {    /* Forward transform */
        /* Calculate wavelet coefficients */
        max_step = (n==0) ? 0 : 1<<n;   /* maximum step */
        for ( step=1 ; step<max_step ; step<<=1 ) {
            if (nX-- > 0) {
                /* Apply forward transform to the rows */
                for ( y=0 ; y<height ; y+=step ) {
                    FLWT1D_Predict ( &Data[y][0], (long)width, 1,
                                     (long)step, (long)N );
                    FLWT1D_Update ( &Data[y][0], (long)width, 1,
                                    (long)step, (long)nTilde, liftingX[nX] );
                }
            }
    
            if (nY-- > 0) {
                /* Apply forward transform to the columns */
                for ( x=0 ; x<width ; x+=step ) {
                    FLWT1D_Predict ( &Data[0][x], (long)height, (long)width,
                                     (long)step, (long)N );
                    FLWT1D_Update ( &Data[0][x], (long)height, (long)width,
                                    (long)step, (long)nTilde, liftingY[nY] );
                }
            }
        }
    } else {   /* Inverse transform */
        /* Switch sign of filter & lifting coefficients */
        for ( i=0 ; i<N*(1+(N>>1)) ; i++ )
            filter[i] = -filter[i];                 /* negate filter */
        for ( x=0 ; x<nX ; x++ )
            for ( k=0 ; k<(width>>1)*nTilde ; k++ )
                liftingX[x][k] = -liftingX[x][k];   /* negate lifting in X */
        for ( y=0 ; y<nY ; y++ )
            for ( k=0 ; k<(height>>1)*nTilde ; k++ )
                liftingY[y][k] = -liftingY[y][k];   /* negate lifting in Y */

        /* Find original dataset */
        max_step = (n==0) ? 0 : 1<<(n-1);   /* initial step */
        for ( step = max_step ; step >= 1 ; step >>= 1 ) {
            if (n <= nY) {
                /* Apply inverse transform to the columns */
                for ( x=0 ; x<width ; x+=step ) {
                    FLWT1D_Update ( &Data[0][x], (long)height, (long)width,
                                    (long)step, (long)nTilde, liftingY[nY-n] );
                    FLWT1D_Predict ( &Data[0][x], (long)height, (long)width,
                                     (long)step, (long)N );
                }
            }
    
            if (n <= nX) {
                /* Apply inverse transform to the rows */
                for ( y=0 ; y<height ; y+=step ) {
                    FLWT1D_Update ( &Data[y][0], (long)width, 1,
                                    (long)step, (long)nTilde, liftingX[nX-n] );
                    FLWT1D_Predict ( &Data[y][0], (long)width, 1,
                                     (long)step, (long)N );
                }
            }
    
            n--;   /* one less level to go */
        }

        /* Switch sign of filter & lifting coefficients again */
        for ( i=0 ; i<N*(1+(N>>1)) ; i++ )
            filter[i] = -filter[i];                 /* negate filter */
        for ( x=0 ; x<nX ; x++ )
            for ( k=0 ; k<(width>>1)*nTilde ; k++ )
                liftingX[x][k] = -liftingX[x][k];   /* negate lifting in X */
        for ( y=0 ; y<nY ; y++ )
            for ( k=0 ; k<(height>>1)*nTilde ; k++ )
                liftingY[y][k] = -liftingY[y][k];   /* negate lifting in Y */
    }
}

/*
 * FLWT1D function: This is a 1D Direct/Inverse Fast Lifted Wavelet
 *                  Trasform. Since the Lifting Scheme is used, the final
 *                  coefficients are organized in-place in the original 
 *                  vector Data[]. In file mallat.c there are
 *                  functions provided to change this organization to
 *                  the Isotropic format originally established by Mallat.
 */
void
FLWT1D ( Vector Data, const int width,
                      const int N, const int nTilde,
                      const int levels, const boolean inverse )
{
    int i, k, x, s,       /* counters */
        step, max_step,   /* step size and max./init. step */
        nX;               /* number of iterations in X and Y directions */
    Flt maxN;             /* maximum number of vanishing moments */


    /* Verify for existence of filter and lifting coefficients */
    if (!filter) {
        Error ("FLWT1D", NO_FILTER_VECTOR, ABORT);
    }

    if ( !liftingX && !liftingY ) {
	fprintf( stderr,
	         "\nFLWT1D(): lifting coefficients are not initialized."
		 "\nUse GetDual() and GetReal() before calling this "
                 "function.\n");
        exit( INPUT_ERROR );
    }

    /* Calculate number of iterations n. It is the maximum  */
    /* value such that the following relation is satisfied. */
    /*        (2^n)*(N-1) <= L < (2^(n+1))*(N-1) + 1        */
    /* Where L = max (signal's length in X-Y direction).    */
    /* and N = max (# dual vanish mom & # real vanish mom)  */
    /* Hence, solving for n, we have the following equation */
    /* for all the cases:  n = floor (log_2((L-1)/(N-1))    */
    maxN = (Flt)MAX(N, nTilde) - (Flt)1;   /* max vanishing moments */
    /* Iterations in X */
    nX = (width == 1) ? 0 : (int)logBaseN( (Flt)(width-1)/maxN, (Flt)2 );
    nX = (nX < 0) ? 0 : MIN(levels, nX);   /* find lowest level */


    if ( !inverse ) {    /* Forward transform */
        /* Calculate wavelet coefficients */
        max_step = (nX==0) ? 0 : 1<<nX;   /* maximum step */
        for ( step=1, s=nX-1 ; step<max_step ; step<<=1, s-- ) {
            /* Apply forward transform to the vector */
            FLWT1D_Predict ( Data, (long)width, 1,
                             (long)step, (long)N );
            FLWT1D_Update ( Data, (long)width, 1,
                            (long)step, (long)nTilde, liftingX[s] );
        }
    } else {   /* Inverse transform */
        /* Switch sign of filter & lifting coefficients */
        for ( i=0 ; i<N*(1+(N>>1)) ; i++ )
            filter[i] = -filter[i];                 /* negate filter */
        for ( x=0 ; x<nX ; x++ )
            for ( k=0 ; k<(width>>1)*nTilde ; k++ )
                liftingX[x][k] = -liftingX[x][k];   /* negate lifting in X */

        /* Find original dataset */
        max_step = (nX==0) ? 0 : 1<<(nX-1);   /* initial step */
        for ( step=max_step, s=0 ; step>=1 ; step>>=1, s++ ) {
            /* Apply inverse transform to the vector */
            FLWT1D_Update ( Data, (long)width, 1,
                            (long)step, (long)nTilde, liftingX[s] );
            FLWT1D_Predict ( Data, (long)width, 1,
                             (long)step, (long)N );
        }

        /* Switch sign of filter & lifting coefficients again */
        for ( i=0 ; i<N*(1+(N>>1)) ; i++ )
            filter[i] = -filter[i];                 /* negate filter */
        for ( x=0 ; x<nX ; x++ )
            for ( k=0 ; k<(width>>1)*nTilde ; k++ )
                liftingX[x][k] = -liftingX[x][k];   /* negate lifting in X */
    }
}

/*
 * FLWT1D_Predict function: The Gamma coefficients are found as an average
 *                          interpolation of their neighbors in order to find
 *                          the "failure to be linear" or "failure to be
 *                          cubic" or the failure to be the order given by
 *                          N-1. This process uses the filter coefficients
 *                          stored in the filter vector and predicts
 *                          the odd samples based on the even ones
 *                          storing the difference in the gammas.
 *                          By doing so, a Dual Wavelet is created.
 */
void
FLWT1D_Predict ( Vector vect, const long size, const long incr,
                              const long step, const long N )
{
    register Vector lambdaPtr,       /* pointer to Lambda coefficients */
                    gammaPtr,        /* pointer to Gamma coefficients */
                    fP, filterPtr;   /* pointers to filter coefficients */
    register long len,               /* number of coefficients at current level */
                  j,                 /* counter for the filter cases */
                  stepIncr;          /* step size between coefficients of the same type */

    long stop1,                      /* number of cases when L < R */
         stop2,                      /* number of cases when L = R */
         stop3,                      /* number of cases when L > R */
         soi;                        /* increment for the middle cases */

    /************************************************/
    /* Calculate values of some important variables */
    /************************************************/
    len       = CEIL(size, step);   /* number of coefficients at current level */
    stepIncr  = (step*incr) << 1;   /* step size betweeen coefficients */

    /************************************************/
    /* Calculate number of iterations for each case */
    /************************************************/
    j     = ODD(len);
    stop1 = N >> 1;
    stop3 = stop1 - j;                /* L > R */
    stop2 = (len >> 1) - N + 1 + j;   /* L = R */
    stop1--;                          /* L < R */

    /***************************************************/
    /* Predict Gamma (wavelet) coefficients (odd guys) */
    /***************************************************/

    filterPtr = filter + N;   /* position filter pointer */

    /* Cases where nbr. left Lambdas < nbr. right Lambdas */
    gammaPtr = vect + (stepIncr >> 1);   /* second coefficient is Gamma */
    while(stop1--) {
        lambdaPtr = vect;   /* first coefficient is always first Lambda */
        j = N;
        do {   /* Gamma update (Gamma - predicted value) */
            *(gammaPtr) -= (*(lambdaPtr)*(*(filterPtr++)));
            lambdaPtr   += stepIncr;   /* jump to next Lambda coefficient */
        } while(--j);   /* use all N filter coefficients */
        /* Go to next Gamma coefficient */
        gammaPtr += stepIncr;
    }

    /* Cases where nbr. left Lambdas = nbr. right Lambdas */
    soi = 0;
    while(stop2--) {
        lambdaPtr = vect + soi;   /* first Lambda to be read */
        fP = filterPtr;   /* filter stays in same values for this cases */
        j = N;
        do {   /* Gamma update (Gamma - predicted value) */
            *(gammaPtr) -= (*(lambdaPtr)*(*(fP++)));
            lambdaPtr   += stepIncr;   /* jump to next Lambda coefficient */
        } while(--j);   /* use all N filter coefficients */
        /* Move start point for the Lambdas */
        soi += stepIncr;
        /* Go to next Gamma coefficient */
        gammaPtr += stepIncr;
    }

    /* Cases where nbr. left Lambdas > nbr. right Lambdas */
    fP = filterPtr;   /* start going backwards with the filter coefficients */
    vect += (soi-stepIncr);   /* first Lambda is always in this place */
    while (stop3--) {
        lambdaPtr = vect;   /* position Lambda pointer */
        j = N;
        do {   /* Gamma update (Gamma - predicted value) */
            *(gammaPtr) -= (*(lambdaPtr)*(*(--fP)));
            lambdaPtr   += stepIncr;   /* jump to next Lambda coefficient */
        } while(--j);   /* use all N filter coefficients */
        /* Go to next Gamma coefficient */
        gammaPtr += stepIncr;
    }
}

/*
 * FLWT1D_Update: the Lambda coefficients have to be "lifted" in order
 *                to find the real wavelet coefficients. The new Lambdas
 *                are obtained  by applying the lifting coeffients stored
 *                in the lifting vector together with the gammas found in
 *                the prediction stage. This process assures that the
 *                moments of the wavelet function are preserved.
 */
void
FLWT1D_Update ( Vector vect, const long size, const long incr,
                             const long step, const long nTilde,
                             const Vector lc )
{
    register Vector lcPtr,   /* pointer to lifting coefficient values */
                    vL,      /* pointer to Lambda values */
                    vG;      /* pointer to Gamma values */
    register long j;         /* counter for the lifting cases */

    long len,                /* number of coefficietns at current level */
         stop1,              /* number of cases when L < R */
         stop2,              /* number of cases when L = R */
         stop3,              /* number of cases when L > R */
         noGammas,           /* number of Gamma coefficients at this level */
         stepIncr,           /* step size between coefficients of the same type */
         soi;                /* increment for the middle cases */


    /************************************************/
    /* Calculate values of some important variables */
    /************************************************/
    len      = CEIL(size, step);   /* number of coefficients at current level */
    stepIncr = (step*incr) << 1;   /* step size between coefficients */
    noGammas = len >> 1 ;          /* number of Gamma coefficients */

    /************************************************/
    /* Calculate number of iterations for each case */
    /************************************************/
    j	  = ODD(len);
    stop1 = nTilde >> 1;
    stop3 = stop1 - j;                   /* L > R */
    stop2 = noGammas - nTilde + 1 + j;   /* L = R */
    stop1--;                             /* L < R */

    /**********************************/
    /* Lift Lambda values (even guys) */
    /**********************************/

    lcPtr = lc;   /* position lifting pointer */

    /* Cases where nbr. left Lambdas < nbr. right Lambdas */
    vG = vect + (stepIncr >> 1);   /* second coefficient is Gamma */
    while(stop1--) {
        vL = vect;   /* lambda starts always in first coefficient */
        j = nTilde;
        do {
            *(vL) += (*(vG)*(*(lcPtr++)));   /* lift Lambda (Lambda + lifting value) */
            vL    += stepIncr;               /* jump to next Lambda coefficient */
        } while(--j);   /* use all nTilde lifting coefficients */
        /* Go to next Gamma coefficient */
        vG += stepIncr;
    }

    /* Cases where nbr. left Lambdas = nbr. right Lambdas */
    soi = 0;
    while(stop2--) {
        vL = vect + soi;   /* first Lambda to be read */
        j = nTilde;
        do {
            *(vL) += (*(vG)*(*(lcPtr++)));   /* lift Lambda (Lambda + lifting value) */
            vL    += stepIncr;               /* jump to next Lambda coefficient */
        } while(--j);   /* use all nTilde lifting coefficients */
        /* Go to next Gamma coefficient */
        vG  += stepIncr;
        /* Move start point for the Lambdas */
        soi += stepIncr;
    }

    /* Cases where nbr. left Lambdas = nbr. right Lambdas */
    vect += (soi - stepIncr);   /* first Lambda is always in this place */
    while(stop3--) {
        vL = vect;   /* position Lambda pointer */
        j = nTilde;
        do {
            *(vL) += (*(vG)*(*(lcPtr++)));   /* lift Lambda (Lambda + lifting value) */
            vL    += stepIncr;               /* jump to next Lambda coefficient */
        } while(--j);   /* use all nTilde lifting coefficients */
        /* Go to next Gamma coefficient */
        vG += stepIncr;
    }
}

/* CODE FOR THE IN-PLACE FLWT (AVERAGE INTERPOLATION CASE, N ODD) */

/*
 * FLWT2D_Haar function: calculates the Haar wavelet coefficients for
 *                       a two dimensional signal. This case corresponds
 *                       to average interpolation with N=1 and wavelets
 *                       with nTilde = 1.
 */
void
FLWT2D_Haar( Matrix Data, const int width, const int height,
                          const int levels, const boolean inverse )
{
    int i, k, x, y, /* counters */
        n,          /* maximum number of levels */
        nX, nY,     /* number of iterations in X and Y directions */
        step,       /* step size of current level */
        max_step;   /* maximum and initial step for the direct and */
                    /* inverse transforms, respectively */
    Flt maxN;       /* maximum number of vanishing moments */

    /* Number of iterations */
    nX = (width==1) ? 0 :
        (int)ceil((Flt)logBaseN((Flt)(width),2.0));
    nX = (nX<0) ? 0 : MIN(nX,levels);
    nY = (height==1) ? 0 :
        (int)ceil((Flt)logBaseN((Flt)(height),2.0));
    nY = (nX<0) ? 0 : MIN(nY,levels);
    n = MAX(nX,nY);


    if ( !inverse ) {    /* Forward transform */
        /* Calculate wavelet coefficients */
        max_step = (n==0) ? 0 : 1<<n;   /* maximum step */
        for ( step = 1 ; step < max_step ; step <<= 1 ) {
            if (nX-- > 0) {
                /* Apply forward transform to the rows */
                for ( y=0 ; y<height ; y+=step ) {
                    FLWT_Haar ( &Data[y][0], (long)width, 1, (long)step );
                }
            }
    
            if (nY-- > 0) {
                /* Apply forward transform to the columns */
                for ( x=0 ; x<width ; x+=step ) {
                    FLWT_Haar ( &Data[0][x], (long)height, (long)width,
                                  (long)step );
                }
            }
        }
    } else {   /* Inverse transform */
        /* Find original dataset */
        max_step = (n==0) ? 0 : 1<<(n-1);   /* initial step */
        for ( step = max_step ; step >= 1 ; step >>= 1 ) {
            if (n <= nY) {
                /* Apply inverse transform to the columns */
                for ( x=0 ; x<width ; x+=step ) {
                    IFLWT_Haar ( &Data[0][x], (long)height, (long)width,
                                   (long)step );
                }
            }
    
            if (n <= nX) {
                /* Apply inverse transform to the rows */
                for ( y=0 ; y<height ; y+=step ) {
                    IFLWT_Haar ( &Data[y][0], (long)width, 1, (long)step );
                }
            }
    
            n--;   /* one less level to go */
        }
    }
}

/*
 * FLWT1D_Haar function: calculates the Haar wavelet coefficients for
 *                       a one dimensional signal. This case corresponds
 *                       to average interpolation with N=1 and wavelets
 *                       with nTilde = 1.
 */
void
FLWT1D_Haar ( Vector Data, const int width,
                           const int levels, const boolean inverse )
{
    int i, k, x, s, /* counters */
        nX;         /* number of iterations in X and Y directions */
    Flt maxN;       /* maximum number of vanishing moments */


    /* Number of iterations */
    nX = (width==1) ? 0 :
        (int)ceil((Flt)logBaseN((Flt)(width),2.0));
    nX = (nX<0) ? 0 : MIN(nX,levels);


    if ( !inverse ) {    /* Forward transform */
        /* Calculate wavelet coefficients */
        for ( s = 0 ; s < nX ; s++ ) {
            /* Apply forward transform to the vector */
            FLWT_Haar ( Data, (long)width, 1, (long)(1<<s) );
        }
    } else {   /* Inverse transform */
        /* Find original dataset */
        for ( s = nX-1 ; s >=0 ; s-- ) {
            /* Apply inverse transform to the vector */
            IFLWT_Haar ( Data, (long)width, 1, (long)(1<<s) ); 
        }
    }
}

/*
 * FLWT_Haar function: finds the Haar and Lambda coefficients for
 *                     the given 1D signal. Boundary cases are not
 *                     considered yet. Thus, all signals must be
 *                     of even lengths.
 */
void
FLWT_Haar ( Vector vect, const long size, const long incr,
                         const long step )
{
    register Vector lambdaPtr,   /* pointer to Lambda coefficients */
                    gammaPtr;    /* pointer to Gamma coefficients */
    register long nL,            /* number of Lambda coefficients */
                  nG,            /* number of Gamma coefficients */
                  stepIncr;      /* step size between coefficients of the same type */
    long len;                    /* number of coefficients at current level */
    Flt  lAve,                   /* average of Lambdas at current level */
         lSum = 0;               /* cumulative of Lambdas before last one */


    /************************************************/
    /* Calculate values of some important variables */
    /************************************************/
    len      = CEIL(size, step);   /* number of coefficients at current level */
    stepIncr = (step*incr) << 1;   /* step size betweeen coefficients */

    /**********************************/
    /* Calculate number of iterations */
    /**********************************/
    nG = (len >> 1);
    nL = len - nG;

    /****************************/
    /* Calculate Lambda average */
    /****************************/
    if ( ODD(len) ) {
        long n = len, s = stepIncr >> 1;
        Flt  t = 0;
        Flt *l = vect;

        while (n--) {
            t += *l;
            l += s;
        }
        lAve = t/(Flt)len;   /* average we've to maintain */
    }

    /**********************************************************/
    /* Calculate Gamma (Haar wavelet) coefficients (odd guys) */
    /* and coarser Lambda coefficients (even guys)            */
    /**********************************************************/
    gammaPtr = vect + (stepIncr >> 1);   /* second coefficient is Gamma */
    lambdaPtr = vect;                    /* first Lambda to be read */
    while (nG--) {
        /* New Gamma (Haar) coefficient */
        *(gammaPtr) -= *(lambdaPtr);
	/* New Lambda coefficient */
        *(lambdaPtr) += (*(gammaPtr)/(Flt)2);
        lSum += *(lambdaPtr);

        /* Go to next Gamma coefficient */
        gammaPtr += stepIncr;
        /* Go to next Lambda coefficient */
        lambdaPtr += stepIncr;
    }


    /*****************************************************/
    /* If ODD(len), then we have one additional case for */
    /* the Lambda calculations. This is a boundary case  */
    /* and, therefore, has different filter values.      */
    /*****************************************************/
    if ( ODD(len) ) {
        *(lambdaPtr) = lAve*(Flt)nL - lSum;
        lSum += *(lambdaPtr);
    }
}

/*
 * IFLWT_Haar function: applies the inverse transform to find Lambda 
 *                      coefficients for the given 1D signal. Boundary 
 *                      cases are not considered yet. Thus, all signals 
 *                      must be of even lengths.
 */
void
IFLWT_Haar( Vector vect, const long size, const long incr,
                         const long step )
{
    register Vector lambdaPtr,   /* pointer to Lambda coefficients */
                    gammaPtr;    /* pointer to Gamma coefficients */
    register long nL,            /* number of Lambda coefficients */
                  nG,            /* number of Gamma coefficients */
                  stepIncr;      /* step size between coefficients of the same type */
    long len;                    /* number of coefficients at current level */
    Flt  lAve,                   /* average of Lambdas at current level */
         lSum = 0;               /* cumulative of Lambdas before last one */


    /************************************************/
    /* Calculate values of some important variables */
    /************************************************/
    len      = CEIL(size, step);   /* number of coefficients at current level */
    stepIncr = (step*incr) << 1;   /* step size betweeen coefficients */

    /**********************************/
    /* Calculate number of iterations */
    /**********************************/
    nG = (len >> 1);
    nL = len - nG;

    /****************************/
    /* Calculate Lambda average */
    /****************************/
    if ( ODD(len) ) {
        long n = nL, s = stepIncr;
        Flt  t = 0;
        Flt *l = vect;

        while (n--) {
            t += *l;
            l += s;
        }
        lAve = t/(Flt)nL;   /* average we've to maintain */
    }


    /**********************************************************/
    /* Calculate Gamma (Haar wavelet) coefficients (odd guys) */
    /* and coarser Lambda coefficients (even guys)            */
    /**********************************************************/
    gammaPtr = vect + (stepIncr >> 1);   /* second coefficient is Gamma */
    lambdaPtr = vect;                    /* first Lambda to be read */
    while (nG--) {
	/* New Lambda coefficient */
        *(lambdaPtr) -= (*(gammaPtr)/(Flt)2);
        lSum += *(lambdaPtr);
	/* New Gamma coefficient */
	*(gammaPtr) += *(lambdaPtr);
        lSum += *(gammaPtr);

        /* Go to next Gamma coefficient */
        gammaPtr += stepIncr;
        /* Go to next Lambda coefficient */
        lambdaPtr += stepIncr;
    }
    
    /*****************************************************/
    /* If ODD(len), then we have one additional case for */
    /* the Lambda calculations. This is a boundary case  */
    /* and, therefore, has different filter values.      */
    /*****************************************************/
    if ( ODD(len) ) {
        *(lambdaPtr) = lAve*(Flt)len - lSum;
        lSum += *(lambdaPtr);
    }
}

/* CODE NEEDED FOR THE FLWT PACKETS */

/*
 * DSFLWT2D function: This function does a disjoint Separable two-
 *                    dimensional Fast Lifted Wavelet Transform on the
 *                    given vector in[] up to one level.
 *                    It reorganizes the coefficients in Mallat format
 *                    and returns the vector out[] with four matrices
 *                    with 4 rows and width*height/4 columns.
 *                    Each row represents a region of the decomposition
 *                    in the following order: upper left for row 0,
 *                    upper right for row 1, bottom left for row 2, and
 *                    bottom right for row 3.
 *
 *                                  Mallat               +---------+ +---------+
 *  +----------------+       +---------+---------+       |         | |         |
 *  |                |       |         |         |       |    0    | |    1    |
 *  |                |       |    0    |    1    |       |         | |         |
 *  |    in-place    |       |         |         |       +---------+ +---------+
 *  |    forward     |  ==>  +---------+---------+  ==>         Separated
 *  |     flwt       |       |         |         |       +---------+ +---------+
 *  |                |       |    2    |    3    |       |         | |         |
 *  |                |       |         |         |       |    2    | |    3    |
 *  +----------------+       +---------+---------+       |         | |         |
 *                                                       +---------+ +---------+
 */
Matrix *
DSFLWT2D ( Matrix in, const int width, const int height,
                      const int N, const int nTilde )
{
    int x, y,                   /* counters */
        x_2 = ( width == 1 ) ? 0 : width>>1,
        y_2 = ( height == 1 ) ? 0 : height>>1;
    Matrix *out;
    Vector outPtr1, outPtr2;

    /* Allocate memory for the ouput regions matrices (4, X/2, Y/2) */
    out = MEM_Init3D( 4, (long)x_2, (long)y_2 );
    if ( !out )
        Error( "DSFLWT2D", MEMORY_MATRIX, ABORT);

    /* Apply forward transform */
    if ( ODD(N) ) {
	FLWT2D_Haar( in, width, height, 1, FALSE );
    } else {
        FLWT2D( in, width, height, N, nTilde, 1, FALSE );
    }

    /* Reorganize coefficients Mallat format */
    for ( y=0 ; y<height ; y++ )
        InplaceChange1D( &in[y][0], width, 1, FALSE );
    for ( x=0 ; x<width ; x++ )
        InplaceChange1D( &in[0][x], height, width, FALSE );
	
    /* Separate regions */
    for ( y=0, outPtr1=&out[0][0][0], outPtr2=&out[1][0][0] ; y<y_2 ; y++ ) {
        for ( x=0 ; x<x_2 ; x++ )
            *(outPtr1++) = in[y][x];   /* fill region 0 */
        for ( x=x_2 ; x<width ; x++ )
            *(outPtr2++) = in[y][x];   /* fill region 1 */
    }
    for ( y=y_2, outPtr1=&out[2][0][0], outPtr2=&out[3][0][0] ; y<height ; y++ ) {
        for ( x=0 ; x<x_2 ; x++ )
            *(outPtr1++) = in[y][x];   /* fill region 2 */
        for ( x=x_2 ; x<width ; x++ )
            *(outPtr2++) = in[y][x];   /* fill region 3 */
    }

    return out;
}


/*
 * ISFLWT2D function: This function does an adjoint Separable two-
 *                    dimensional Fast Lifted Wavelet Transform on the
 *                    given matrix in[][] up to one level.
 *                    It takes the vectors in in[] to create an initial
 *                    matrix out[][] of with Mallat format and size
 *                    width*height. Then, it reorganizes it in Lifiting
 *                    format and applies the inverse transform. The result
 *                    is the in-place version of the inverse transform.
 *
 *                                  Mallat               +---------+ +---------+
 *  +----------------+       +---------+---------+       |         | |         |
 *  |                |       |         |         |       |    0    | |    1    |
 *  |                |       |    0    |    1    |       |         | |         |
 *  |    in-place    |       |         |         |       +---------+ +---------+
 *  |    inverse     |  <==  +---------+---------+  <==         Separated
 *  |      flwt      |       |         |         |       +---------+ +---------+
 *  |                |       |    2    |    3    |       |         | |         |
 *  |                |       |         |         |       |    2    | |    3    |
 *  +----------------+       +---------+---------+       |         | |         |
 *                                                       +---------+ +---------+
 */
Matrix
ISFLWT2D ( Matrix *in, const int width, const int height, 
                       const int N, const int nTilde )
{
    int x, y,                /* counters */
        x_2 = width>>1,
        y_2 = height>>1;
    Matrix out;
    Vector inPtr1, inPtr2;

    /* Allocate memory for output coefficients matrix (width, height) */
    out = MEM_Init2D( (long)width, (long)height );
    if ( !out )
        Error( "SFLWT2D", MEMORY_MATRIX, ABORT);

    /* Adjoint regions */
    for ( y=0, inPtr1=&in[0][0][0], inPtr2=&in[1][0][0] ; y<y_2 ; y++ ) {
        for ( x=0 ; x<x_2 ; x++ )
            out[y][x] = *(inPtr1++);   /* fill region 0 */
        for ( x=x_2 ; x<width ; x++ )
            out[y][x] = *(inPtr2++);   /* fill region 1 */
    }
    for ( y=y_2, inPtr1=&in[2][0][0], inPtr2=&in[3][0][0] ; y<height ; y++ ) {
        for ( x=0 ; x<x_2 ; x++ )
            out[y][x] = *(inPtr1++);   /* fill region 2 */
        for ( x=x_2 ; x<width ; x++ )
            out[y][x] = *(inPtr2++);   /* fill region 3 */
    }

    /* Reorganize coefficients in Lifting format */
    for ( x=0 ; x<width ; x++ )
        InplaceChange1D( &out[0][x], height, width, TRUE );
    for ( y=0 ; y<height ; y++ )
        InplaceChange1D( &out[y][0], width, 1, TRUE );

    /* Apply in-place inverse transform */
    if ( ODD(N) ) {
	FLWT2D_Haar( out, width, height, 1, TRUE );
    } else {
        FLWT2D( out, width, height, N, nTilde, 1, TRUE );
    }

    return out;
}

/*
 * IFLWTP2D function: This function does an inverse two-dimensional
 *                    Fast Lifted Wavelet Transform on the given matrix
 *                    in[][] up to one level.
 *                    The matrix in[][] is in Mallat format with size
 *                    width*height. The routine reorganizes it in Lifiting
 *                    format and applies the inverse transform. The result
 *                    is the in-place version of the inverse transform.
 *
 *                                  Mallat        
 *  +----------------+       +---------+---------+
 *  |                |       |         |         |
 *  |                |       |    0    |    1    |
 *  |    in-place    |       |         |         |
 *  |    inverse     |  <==  +---------+---------+
 *  |      flwt      |       |         |         |
 *  |                |       |    2    |    3    |
 *  |                |       |         |         |
 *  +----------------+       +---------+---------+
 */
void
IFLWTP2D ( Matrix in, const int width, const int height,
                      const int N, const int nTilde )
{
    int x, y;   /* counters */

    /* Reorganize coefficients in Lifting format */
    for ( x=0 ; x<width ; x++ )
        InplaceChange1D( &in[0][x], height, width, TRUE );
    for ( y=0 ; y<height ; y++ )
        InplaceChange1D( &in[y][0], width, 1, TRUE );

    /* Apply in-place inverse transform */
    if ( ODD(N) ) {
	FLWT2D_Haar( in, width, height, 1, TRUE );
    } else {
        FLWT2D( in, width, height, N, nTilde, 1, TRUE );
    }
}

/*
 * DSFLWT1D function: This function does a disjoint Separable one-
 *                    dimensional Fast Lifted Wavelet Transform on the
 *                    given vector in up to one level.
 *                    It reorganizes the coefficients in Mallat format
 *                    and returns the vector out[] with two vectors
 *                    with width/2 columns.
 *
 *       In-Place                 Mallat               
 *  +----------------+     +---------+---------+     +---------+ +---------+
 *  |    (forward)   | ==> |         |         | ==> |         | | |
 *  +----------------+     +---------+---------+     +---------+ +---------+
 */
Vector *
DSFLWT1D ( Vector in, const int width, const int N, const int nTilde )
{
    int x,                   /* counter */
        x_2 = ( width == 1 ) ? 0 : width>>1;
    Vector *out;

    /* Allocate memory for the ouput regions vectors (2, X/2) */
    out = MEM_InitVecs( 2, (long)x_2 );
    if ( !out )
        Error( "DSFLWT1D", MEMORY_MATRIX, ABORT);

    /* Apply forward transform */
    FLWT1D( in, width, N, nTilde, 1, FALSE );

    /* Reorganize coefficients Mallat format */
    InplaceChange1D( in, width, 1, FALSE );
	
    /* Separate regions */
    for ( x=0 ; x<x_2 ; x++ ) {
        out[0][x] = in[x];   /* fill region 0 */
        out[1][x] = in[x+x_2];   /* fill region 1 */
    }

    return out;
}


/*
 * ISFLWT1D function: This function does an adjoint Separable one-
 *                    dimensional Fast Lifted Wavelet Transform on the
 *                    given vectors in[] up to one level.
 *                    It takes the vectors in in to create an initial
 *                    vector out[] of with Mallat format and size
 *                    width. Then, it reorganizes it in Lifting
 *                    format and applies the inverse transform. The
 *                    result is the in-place version of the inverse 
 *                    transform.
 *
 *       In-Place                 Mallat               
 *  +----------------+     +---------+---------+     +---------+ +---------+
 *  |    (inverse)   | <== |         |         | <== |         | | |
 *  +----------------+     +---------+---------+     +---------+ +---------+
 */
Vector
ISFLWT1D ( Vector *in, const int width, const int N, const int nTilde )
{
    int x,                /* counter */
        x_2 = width>>1;
    Vector out;

    /* Allocate memory for output coefficients vector (width) */
    out = vector( 0, (long)(width-1) );
    if ( !out )
        Error( "ISFLWT1D", MEMORY_MATRIX, ABORT);

    /* Adjoint regions */
    for ( x=0 ; x<x_2 ; x++ ) {
        out[x] = in[0][x];   /* fill region 0 */
        out[x+x_2] = in[1][x];   /* fill region 1 */
    }

    /* Reorganize coefficients in Lifting format */
    InplaceChange1D( out, width, 1, TRUE );

    /* Apply in-place inverse transform */
    FLWT1D( out, width, N, nTilde, 1, TRUE );

    return out;
}

/*
 * IFLWTP1D function: This function does an inverse one-dimensional
 *                    Fast Lifted Wavelet Transform on the given vector
 *                    in[] up to one level.
 *                    The vector in[] is in Mallat format with size
 *                    width. The routine reorganizes it in Lifting
 *                    format and applies the inverse transform. The
 *                    result is the in-place version of the inverse 
 *                    transform.
 *
 *       In-Place                   Mallat        
 *  +----------------+       +---------+---------+
 *  |    (inverse)   |  <==  |         |         |
 *  +----------------+       +---------+---------+
 */
void
IFLWTP1D ( Vector in, const int width, const int N, const int nTilde )
{
    /* Reorganize coefficients in Lifting format */
    InplaceChange1D( in, width, 1, TRUE );

    /* Apply in-place inverse transform */
    FLWT1D( in, width, N, nTilde, 1, TRUE );
}

